/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */



#include <Install.h>

#include <xosldata.h>
#include <mem.h>
#include <disk.h>


CInstaller::CInstaller(CTextUI &TextUIToUse, CPartList &PartListToUse):
	TextUI(TextUIToUse),
	PartList(PartListToUse),
	FatInstall(TextUIToUse,XoslFiles,DosFile),
	FsCreator(TextUIToUse,XoslFiles,DosFile),
	UpgradeData(TextUIToUse,XoslFiles,DosFile,PartListToUse)
{
}

CInstaller::~CInstaller()
{
}

int CInstaller::Install(CVesa::TGraphicsMode GraphicsMode, CMouse::TMouseType MouseType, const CDosDriveList::CDosDrive &DosDrive)
{
	TIPL Ipl;

	if (CreateXoslData(GraphicsMode,MouseType) == -1)
		return -1;
	if (CreateBootItem() == -1)
		return -1;
	if (CreateLastConf() == -1)
		return -1;
	CheckBackupSpaceAvail();
	BackupMbr();

	if (FatInstall.CreateIpl(DosDrive,Ipl) == -1)
		return -1;
	if (FatInstall.InstallFiles(DosDrive) == -1)
		return -1;
	if (FatInstall.InstallIpl(Ipl,BackupSpace) == -1)
		return -1;

	TextUI.OutputStr("\nInstall complete\n");
	return 0;
}

int CInstaller::Install(CVesa::TGraphicsMode GraphicsMode, CMouse::TMouseType MouseType, int PartIndex)
{
	const TPartition *Partition;
	TIPL Ipl;
	CDosDriveList::CDosDrive DosDrive;

	Partition = PartList.GetPartition(PartIndex);
	DosDrive.Drive = Partition->Drive;
	DosDrive.FATType = FATTYPE_FAT16; // dedicated partition always FAT16
	DosDrive.StartSector = Partition->StartSector;

	if (Partition->SectorCount < 600) {
		TextUI.OutputStr("XOSL 1.1.0 requires a partition of\nat least 300kb\n\n");
		return -1;
	}
		
	if (CreateXoslData(GraphicsMode,MouseType) == -1)
		return -1;
	if (CreateBootItem() == -1)
		return -1;
	if (CreateLastConf() == -1)
		return -1;
	CheckBackupSpaceAvail();
	BackupMbr();
	
	if (FsCreator.InstallFs(Partition->Drive,Partition->StartSector) == -1)
		return -1;

	if (FatInstall.CreateIpl(DosDrive,Ipl) == -1)
		return -1;

	TextUI.OutputStr("Updating partition table...");
	PartList.SetFsType(PartIndex,0x78);
	PartList.WriteStructure();
	TextUI.OutputStr("done\n");


	if (FatInstall.InstallIpl(Ipl,BackupSpace) == -1)
		return -1;

	TextUI.OutputStr("\nInstall complete\n");
	return 0;
}

int CInstaller::Uninstall(const CDosDriveList::CDosDrive &DosDrive)
{
	if (RestoreMbr() == -1)
		return -1;
	FatInstall.RemoveXoslFiles(DosDrive.DriveChar);
	TextUI.OutputStr("Uninstall complete\n");
	return 0;
}

int CInstaller::Uninstall(int PartIndex)
{
	const TPartition *Partition;

	if (RestoreMbr() == -1)
		return -1;
	Partition = PartList.GetPartition(PartIndex);
	FsCreator.RestorePartition(Partition->Drive,Partition->StartSector);
	TextUI.OutputStr("Uninstall complete\n");
	return 0;
}

int CInstaller::Upgrade(const CDosDriveList::CDosDrive &DosDrive)
{
	TIPL Ipl;

	if (UpgradeData.CreateDataFiles(DosDrive.DriveChar) == -1)
		return -1;

	CheckBackupSpaceAvail();

	if (FatInstall.CreateIpl(DosDrive,Ipl) == -1)
		return -1;
	if (FatInstall.InstallFiles(DosDrive) == -1)
		return -1;
	if (FatInstall.InstallIpl(Ipl,BackupSpace) == -1)
		return -1;

	TextUI.OutputStr("\nUpgrade complete\n");
	return 0;
}

int CInstaller::Upgrade(int PartIndex)
{
	const TPartition *Partition;
	TIPL Ipl;
	CDosDriveList::CDosDrive DosDrive;

	Partition = PartList.GetPartition(PartIndex);
	DosDrive.Drive = Partition->Drive;
	DosDrive.FATType = FATTYPE_FAT16; // dedicated partition always FAT16
	DosDrive.StartSector = Partition->StartSector;

	if (Partition->SectorCount < 600) {
		TextUI.OutputStr("XOSL 1.1.0 requires a partition of\nat least 300kb\n\n");
		return -1;
	}
		
	if (UpgradeData.CreateDataFiles(Partition) == -1)
		return -1;

	CheckBackupSpaceAvail();
	
	if (FsCreator.InstallFs(Partition->Drive,Partition->StartSector) == -1)
		return -1;

	if (FatInstall.CreateIpl(DosDrive,Ipl) == -1)
		return -1;

	TextUI.OutputStr("Updating partition table...");
	PartList.SetFsType(PartIndex,0x78);
	PartList.WriteStructure();
	TextUI.OutputStr("done\n");

	if (FatInstall.InstallIpl(Ipl,BackupSpace) == -1)
		return -1;

	TextUI.OutputStr("\nUpgrade complete\n");
	return 0;
}




int CInstaller::CreateXoslData(CVesa::TGraphicsMode GraphicsMode, CMouse::TMouseType MouseType)
{
	CXoslData XoslData;
	CXoslData::CGraphData &GraphData = XoslData.GraphData;
	CXoslData::CColorData &ColorData = XoslData.ColorData;
	CXoslData::CMouseData &MouseData = XoslData.MouseData;
	CXoslData::CMiscPref &MiscPref = XoslData.MiscPref;
	int fh;

	TextUI.OutputStr("Creating data file...");

	memset(&XoslData,0,sizeof (CXoslData));

	GraphData.VideoMode = GraphicsMode;
	GraphData.FrameMove = 1;
	
	ColorData.ColorScheme = 15; // XOSL default
	ColorData.FadeColorInt = 100; // 100% -> White

	MouseData.MouseType = MouseType;
	MouseData.MouseSpeed = 5;
	MouseData.PS2Sampling = 5;

	MiscPref.KeyRestart = 0x4400;    // F10
	MiscPref.KeyColdReboot = 0x7100; // Alt-F10
	MiscPref.KeyShutdown = 0x6700;   // Ctrl-F10
	MiscPref.KeyCycle = 0x1100;      // Alt-W
	MiscPref.ActiveAllow = 1;

	XoslData.XoslVersion = 2;

	if ((fh = DosFile.Create(XoslFiles.GetXoslDataName())) == -1) {
		TextUI.OutputStr("failed\nUnable to create %s\n",XoslFiles.GetXoslDataName());
		return -1;
	}
	DosFile.Write(fh,&XoslData,sizeof (CXoslData));
	DosFile.Close(fh);
	TextUI.OutputStr("done\n");


	return 0;
}


int CInstaller::CreateBootItem()
{
	char BootItemData[BOOTITEM_FILESIZE];
	int hFile;
	
	TextUI.OutputStr("Creating boot items file...");
	memset(BootItemData,0,BOOTITEM_FILESIZE);
	if ((hFile = DosFile.Create(XoslFiles.GetBootItemName())) == -1) {
		TextUI.OutputStr("failed\nUnable to create %s\n",XoslFiles.GetBootItemName());
		return -1;
	}
	DosFile.Write(hFile,BootItemData,BOOTITEM_FILESIZE);
	DosFile.Close(hFile);
	TextUI.OutputStr("done\n");
	return 0;
}

int CInstaller::CreateLastConf()
{
	char LastConfData[LASTCONF_FILESIZE];
	int hFile;
	
	TextUI.OutputStr("Creating config file...");
	memset(LastConfData,0,LASTCONF_FILESIZE);
	if ((hFile = DosFile.Create(XoslFiles.GetLastConfName())) == -1) {
		TextUI.OutputStr("failed\nUnable to create %s\n",XoslFiles.GetLastConfName());
		return -1;
	}
	DosFile.Write(hFile,LastConfData,LASTCONF_FILESIZE);
	DosFile.Close(hFile);
	TextUI.OutputStr("done\n");
	return 0;
}

void CInstaller::CheckBackupSpaceAvail()
{
	const TPartition *Partition;
	int Index, Count;

	Count = PartList.GetCount();
	for (Index = 0; Index < Count; ++Index) {
		Partition = PartList.GetPartition(Index);
		if (Partition->Drive == 0x80 && Partition->StartSector < 5) {
			BackupSpace = 0;
			return;
		}
	}
	BackupSpace = 1;

}

void CInstaller::BackupMbr()
{
	CDisk Disk;
	CMbr Mbr;
	int Index;
	unsigned short Checksum;

	TextUI.OutputStr("Creating backup of MBR...");
	if (!BackupSpace) {
		TextUI.OutputStr("ignored\nNo room for backup\n");
		return;
	}
	Disk.Map(0x80,0);
	Disk.Lock();
	Disk.Read(0,&Mbr,1);
	for (Checksum = Index = 0; Index < 223; ++Index)
		Checksum += Mbr.IPL[Index];

	Mbr.MagicNumber = Checksum;
	Mbr.XoslSignature = 0x4c534f58;

	// Backup of original MBR
	Disk.Write(1,&Mbr,1);
	Disk.Unlock();
	TextUI.OutputStr("done\n");
}


int CInstaller::RestoreMbr()
{
	CDisk Disk;
	CMbr MbrSrc, MbrDest;
	int Index;
	unsigned short Checksum;
	int hFile;

	TextUI.OutputStr("Locating backup of MBR...");

	Disk.Map(0x80,0);
	Disk.Lock();
	Disk.Read(1,&MbrSrc,1);

	for (Checksum = Index = 0; Index < 223; ++Index)
		Checksum += MbrSrc.IPL[Index];


	if (MbrSrc.MagicNumber == Checksum && MbrSrc.XoslSignature == 0x4c534f58)
		TextUI.OutputStr("found\n");
	else {
		TextUI.OutputStr("failed\nUsing default MBR instead\n");
		if ((hFile = DosFile.Open("DEFAULT.MBR",CDosFile::accessReadOnly)) == -1) {
			TextUI.OutputStr("Unable to open DEFAULT.MBR\n");
			return -1;
		}
		DosFile.Read(hFile,&MbrSrc,512);
		DosFile.Close(hFile);
	}

	TextUI.OutputStr("Restoring MBR...");
	Disk.Read(0,&MbrDest,1);
	memcpy(MbrDest.IPL,MbrSrc.IPL,446);
	Disk.Write(0,&MbrDest,1);
	Disk.Unlock();
	TextUI.OutputStr("done\n");
	return 0;
}

