/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */


#include <Upgrade.h>

#include <xosldata.h>
#include <datastr.h>
#include <mem.h>
#include <string.h>
#include <fat16.h>
#include <disk.h>

CUpgradeData::CUpgradeData(CTextUI &TextUIToUse, CXoslFiles &XoslFilesToUse, CDosFile &DosFileToUse, CPartList &PartListToUse):
	TextUI(TextUIToUse),
	XoslFiles(XoslFilesToUse),
	DosFile(DosFileToUse),
	PartList(PartListToUse)
{

}

CUpgradeData::~CUpgradeData()
{

}


CUpgradeData::CreateDataFiles(char DriveNumber)
{
	if (LoadDataFiles(DriveNumber) == -1)
		return -1;
	if (ConvertData() == -1)
		return -1;
	return 0;
}

CUpgradeData::CreateDataFiles(const TPartition *Partition)
{
	CFAT16 FileSystem; // dedicated partition always FAT16
	CDisk Disk;
	TBootFAT16 BootRecord;
	
	// make sure it's XOSL on a _dedicated_ partition, and not
	// XOSL on a FAT16 partition
	Disk.Map(Partition->Drive,Partition->StartSector);
	Disk.Read(0,&BootRecord,1);
	if (memcmp(BootRecord.OEM_ID,"XOSLINST",8) != 0 && memcmp(BootRecord.OEM_ID,"GEURTVOS",8) != 0) {
		TextUI.OutputStr("Partition is not valid\n");
		return -1;
	}

	if (FileSystem.Mount(Partition->Drive,Partition->StartSector) == -1) {
		TextUI.OutputStr("Unable to mount partition\n");
		return -1;
	}

	if (LoadDataFiles(FileSystem) == -1)
		return -1;
	if (ConvertData() == -1)
		return -1;
	return 0;
}



int CUpgradeData::LoadDataFiles(char DriveNumber)
{
	char FileName[32];
	TextUI.OutputStr("Loading data files...");

	FileName[0] = DriveNumber;
	FileName[1] = ':';
	FileName[2] = '\\';

	strcpy(&FileName[3],XoslFiles.GetXoslDataName());
	if (LoadFile(FileName,XOSL_DATA_SIZE,XoslDataBuffer) == -1)
		return -1;

	// check what version is installed.
	// 1 -> 1.0.0 or 1.1.0.dv1; 2 -> 1.1.0

	InstalledVersion = *(int *)XoslDataBuffer;
	if (InstalledVersion != 1 && InstalledVersion != 2) {
		TextUI.OutputStr("failed\nUnrecognized version\n");
		return -1;
	}

	strcpy(&FileName[3],XoslFiles.GetBootItemName());
	if (LoadFile(FileName,BOOT_ITEM_SIZE,BootItemBuffer) == -1)
		return -1;
	strcpy(&FileName[3],XoslFiles.GetLastConfName());
	if (LoadFile(FileName,LAST_CONF_SIZE_OLD,LastConfBuffer) == -1)
		if (LoadFile(FileName,LAST_CONF_SIZE_NEW,LastConfBuffer) == -1) {
			TextUI.OutputStr("failed\n%s: Unrecognized version\n",FileName);
			return -1;
		}
	TextUI.OutputStr("done\n");
	return 0;
}
	


int CUpgradeData::LoadFile(const char *FileName, int FileSize, void *Buffer)
{
	int hFile;
	long Size;

	Size = DosFile.FileSize(FileName);

	if (Size == -1 || Size != FileSize)
		return -1;

	hFile = DosFile.Open(FileName,CDosFile::accessReadOnly);
	DosFile.Read(hFile,Buffer,FileSize);
	DosFile.Close(hFile);
	return 0;
}



int CUpgradeData::LoadDataFiles(CFileSystem &FileSystem)
{
	char FileName[11];
	unsigned short Size;

	TextUI.OutputStr("Loading data files...");
	GetFat16FileName(FileName,XoslFiles.GetXoslDataName());
	if (FileSystem.ReadFile(FileName,XoslDataBuffer) != XOSL_DATA_SIZE) {
		TextUI.OutputStr("failed\nUnable to open %s\n",FileName);
		return -1;
	}

	// check what version is installed.
	// 1 -> 1.0.0 or 1.1.0; 2 -> 1.1.0
	InstalledVersion = *(int *)XoslDataBuffer;
	if (InstalledVersion != 1 && InstalledVersion != 2) {
		TextUI.OutputStr("failed\nUnrecognized version\n");
		return -1;
	}

	GetFat16FileName(FileName,XoslFiles.GetBootItemName());
	if (FileSystem.ReadFile(FileName,BootItemBuffer) != BOOT_ITEM_SIZE) {
		TextUI.OutputStr("failed\nUnable to open %s\n",FileName);
		return -1;
	}

	GetFat16FileName(FileName,XoslFiles.GetLastConfName());
	Size = FileSystem.ReadFile(FileName,LastConfBuffer);
	if (Size != LAST_CONF_SIZE_OLD && Size != LAST_CONF_SIZE_NEW) {
		TextUI.OutputStr("failed\nUnable to open %s\n",FileName);
		return -1;
	}
	TextUI.OutputStr("done\n");
	return 0;
}

int CUpgradeData::ConvertData()
{
	CXoslData::CMiscPref *MiscPref;

	if (InstalledVersion == 2)
		return Upgrade110Data();
	MiscPref = &((CXoslData *)XoslDataBuffer)->MiscPref;
	if (!MiscPref->KeyColdReboot && !MiscPref->KeyRestart && !MiscPref->KeyShutdown)
		return Upgrade100Data();
	return Upgrade110dv1Data();
}

int CUpgradeData::Upgrade110Data()
{
	TextUI.OutputStr("XOSL 1.1.0 found\n");

	// just save the loaded data files
	return SaveDataFiles();
}


int CUpgradeData::Upgrade100Data()
{
	CXoslData &XoslData = *(CXoslData *)XoslDataBuffer;
	CXosl100DataStructs::CItemFile &ItemFile = *(CXosl100DataStructs::CItemFile *)BootItemBuffer;
//	CXosl100DataStructs::CLastConfig &LastConfig = *(CXosl100DataStructs::CLastConfig *)LastConfBuffer;

	CXosl110DataStructs::CItemFile NewItemFile;
//	CXosl110DataStructs::CLastConfig NewLastConfig;

	CXosl100DataStructs::CItemFile::CBootItem *BootItem;
	CXosl110DataStructs::CItemFile::CBootItem *NewBootItem;

//	CXosl100DataStructs::CLastConfig::CConfPart *LastPart;
//	CXosl110DataStructs::CLastConfig::CPartKey *NewLastPart;

	int Index;
/*	long Cylinder, Head, Sector;
	long HeadCount, SectorCount;*/

	CXosl110DataStructs::CLastConfig &NewLastConfig = *(CXosl110DataStructs::CLastConfig *)LastConfBuffer;
	CXosl110DataStructs::CLastConfig::CPartKey *KeyList;
	int Count;
	const TPartition *Partition;
	int Skip;


	TextUI.OutputStr("XOSL 1.0.0 found\n");
	TextUI.OutputStr("Converting data...");

	memset(&NewItemFile,0,sizeof (CXosl110DataStructs::CItemFile));
//	memset(&NewLastConfig,0,sizeof (CXosl110DataStructs::CLastConfig));
	memset(&NewLastConfig,0,LAST_CONF_SIZE_NEW);

	// convert boot items 
	NewItemFile.BootItemCount = ItemFile.BootItemCount;
	NewItemFile.PartCount = ItemFile.PartCount;
	NewItemFile.InstantBoot = ItemFile.InstantBoot;
	NewItemFile.Timeout = ItemFile.Timeout;

	for (Index = 0; Index < NewItemFile.BootItemCount; ++Index) {
		BootItem = &ItemFile.BootItems[Index];
		NewBootItem = &NewItemFile.BootItems[Index];

		memcpy(NewBootItem->ItemName,BootItem->ItemName,48);
		memcpy(NewBootItem->Keys,BootItem->Keys,16);
		NewBootItem->Activate = BootItem->Activate;
		NewBootItem->Password = BootItem->Password;
		NewBootItem->PartIndex = BootItem->PartIndex;
		NewBootItem->Valid = BootItem->Valid;
		NewBootItem->Hotkey = 0;
		NewBootItem->FixDriveNum = 0;
		memcpy(NewBootItem->HidingItems,BootItem->HidingItems,92);
	}
	memcpy(BootItemBuffer,&NewItemFile,sizeof (CXosl110DataStructs::CItemFile));

	// converting last config fails, and because I'm too fed-up with it, instead
	// of converting, a new lastconf.xdf is created here.

	NewLastConfig.PartitionCount = Count = PartList.GetCount();
	KeyList = NewLastConfig.LastPartList;
	for (Index = 0; Index < Count; ++Index) {
		Partition = PartList.GetPartition(Index);
		if (Partition->Type == PART_PRIMARY)
			switch (Partition->FSType) {
				case 0x0f: // Extended LBA
				case 0x1f: // Hidden Extended LBA
				case 0x85: // Linux Extended
					Skip = 1;
					break;
				default:
					Skip = 0;
					break;
			}
		if (!Skip) {
			KeyList->Drive = Partition->Drive;
			KeyList->StartSector = Partition->StartSector;
			++KeyList;
		}
	}


/*	// convert last config
	NewLastConfig.PartitionCount = LastConfig.PartitionCount;

	for (Index = 0; Index < NewLastConfig.PartitionCount; ++Index) {
		LastPart = &LastConfig.Partitions[Index];
		NewLastPart = &NewLastConfig.LastPartList[Index];

		NewLastPart->Drive = LastPart->DrvHead & 0x00ff;

		Head = LastPart->DrvHead >> 8;
		Sector = (LastPart->SectCyl & 0x3f) - 1;
		Cylinder = (LastPart->SectCyl >> 8) | ((LastPart->SectCyl & 0xc0) << 2);

		HeadCount = (long)LastPart->PTabEntry.EndHead + 1;
		SectorCount = LastPart->PTabEntry.EndSectCyl & 0x3f;

		NewLastPart->StartSector = ((Cylinder * HeadCount) + Head) * SectorCount + Sector;


	}
	memcpy(LastConfBuffer,&NewLastConfig,sizeof(CXosl110DataStructs::CLastConfig));*/




	// update xosl data
	XoslData.GraphData.VideoMode = (CVesa::TGraphicsMode)((XoslData.GraphData.VideoMode - 0x0101) / 2 + 1);
	XoslData.ColorData.FadeColorInt = 100; // 100% -> White

	XoslData.MiscPref.KeyRestart = 0x4400;    // F10
	XoslData.MiscPref.KeyColdReboot = 0x7100; // Alt-F10
	XoslData.MiscPref.KeyShutdown = 0x6700;   // Ctrl-F10
	XoslData.MiscPref.KeyCycle = 0x1100;      // Alt-W
	XoslData.MiscPref.ActiveAllow = 1;

	XoslData.XoslVersion = 2;

	TextUI.OutputStr("done\n");
	return SaveDataFiles();
}

int CUpgradeData::Upgrade110dv1Data()
{
	CXoslData &XoslData = *(CXoslData *)XoslDataBuffer;
	CXosl100DataStructs::CItemFile &ItemFile = *(CXosl100DataStructs::CItemFile *)BootItemBuffer;

	CXosl110DataStructs::CItemFile NewItemFile;

	CXosl100DataStructs::CItemFile::CBootItem *BootItem;
	CXosl110DataStructs::CItemFile::CBootItem *NewBootItem;

	int Index;

	TextUI.OutputStr("XOSL 1.1.0.dv1 found\n");
	TextUI.OutputStr("Converting data...");


	memset(&NewItemFile,0,sizeof (CXosl110DataStructs::CItemFile));

	// convert boot items 
	NewItemFile.BootItemCount = ItemFile.BootItemCount;
	NewItemFile.PartCount = ItemFile.PartCount;
	NewItemFile.InstantBoot = ItemFile.InstantBoot;
	NewItemFile.Timeout = ItemFile.Timeout;

	for (Index = 0; Index < NewItemFile.BootItemCount; ++Index) {
		BootItem = &ItemFile.BootItems[Index];
		NewBootItem = &NewItemFile.BootItems[Index];

		memcpy(NewBootItem->ItemName,BootItem->ItemName,48);
		memcpy(NewBootItem->Keys,BootItem->Keys,16);
		NewBootItem->Activate = BootItem->Activate;
		NewBootItem->Password = BootItem->Password;
		NewBootItem->PartIndex = BootItem->PartIndex;
		NewBootItem->Valid = BootItem->Valid;
		NewBootItem->Hotkey = 0;
		NewBootItem->FixDriveNum = 0;
		memcpy(NewBootItem->HidingItems,BootItem->HidingItems,92);
	}
	memcpy(BootItemBuffer,&NewItemFile,sizeof (CXosl110DataStructs::CItemFile));

	// update xosl data
	XoslData.GraphData.VideoMode = (CVesa::TGraphicsMode)((XoslData.GraphData.VideoMode - 0x0101) / 2 + 1);

	XoslData.MiscPref.KeyRestart = 0x4400;    // F10
	XoslData.MiscPref.KeyColdReboot = 0x7100; // Alt-F10
	XoslData.MiscPref.KeyShutdown = 0x6700;   // Ctrl-F10
	XoslData.MiscPref.KeyCycle = 0x1100;      // Alt-W
	XoslData.MiscPref.ActiveAllow = 1;

	XoslData.XoslVersion = 2;

	TextUI.OutputStr("done\n");
	return SaveDataFiles();
}


int CUpgradeData::SaveDataFiles()
{
	const char *FileName;

	FileName = XoslFiles.GetXoslDataName();
	TextUI.OutputStr("Creating %s...",FileName);
	if (SaveFile(FileName,XOSL_DATA_SIZE,XoslDataBuffer) == -1)
		return -1;

	FileName = XoslFiles.GetBootItemName();
	TextUI.OutputStr("done\nCreating %s...",FileName);
	if (SaveFile(FileName,BOOT_ITEM_SIZE,BootItemBuffer) == -1)
		return -1;

	FileName = XoslFiles.GetLastConfName();
	TextUI.OutputStr("done\nCreating %s...",FileName);
	if (SaveFile(FileName,LAST_CONF_SIZE_NEW,LastConfBuffer) == -1)
		return -1;
	TextUI.OutputStr("done\n");
	return 0;
}

int CUpgradeData::SaveFile(const char *FileName, int FileSize, void *Buffer)
{
	int hFile;

	if ((hFile = DosFile.Create(FileName)) == -1) {
		TextUI.OutputStr("failed\nUnable to create %s\n",FileName);
		return -1;
	}
	DosFile.Write(hFile,Buffer,FileSize);
	DosFile.Close(hFile);
	return 0;
}


void CUpgradeData::GetFat16FileName(char *Fat16Name, const char *DosName)
{
	/*
	 * assumtions:
	 * - all files have a 3-char extension
	 */

	int Length = strlen(DosName);

	memset(Fat16Name,' ',8 + 3);
	memcpy(Fat16Name,DosName,Length - 4); // DosName - ".xxx"
	memcpy(&Fat16Name[8],&DosName[Length - 3],3);
}
