/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <graph.h>
#include <mem.h>
#include <int10.h>
#include <vpgraph.h>
#include <mouse.h>
#include <retrace.h>
#include <cursor.h>
#include <pmgraph.h>

#define CURSOR_WIDTH 11
#define CURSOR_HEIGHT 19

typedef struct SGraphQueue{
	long PhysParams;
	long far *Params;
	long Func;
	long PhysNext;
	struct SGraphQueue far *Next;
} TGraphQueue, far *PGraphQueue;

extern "C" {
	void InitViewport(long Width, long Height);
	void InitVBE2(unsigned long Addr);
	int ConfirmLFB(unsigned long Addr);
};

static struct {
	int VesaMode;
	int Width, Height;
	int LFBSupport;
	unsigned long LinearBuffer;
	int Granularity;
	unsigned long SwitchAddr;
	const char *ModeName;
} VesaModeList[] = {
	{0x0003,0,0,0,0,0,0,"Text 80x25"},
	{-1,640,480,0,0,0,0,"640x480"},
	{-1,800,600,0,0,0,0,"800x600"},
	{-1,1024,768,0,0,0,0,"1024x768"},
	{-1,1280,1024,0,0,0,0,"1280x1024"},
	{-1,1600,1200,0,0,0,0,"1600x1200"}
};


long GRWidth, GRHeight;
int GRThisBank;
long TextHeight;

static long TempVPLeft, TempVPTop;

TGraphQueue GraphQueue = {0,0,0,NULL};
PGraphQueue Tail = &GraphQueue;

CGraph *Graph;

CGraph::CGraph(CMouse *MouseToUse)
{
	Mouse = MouseToUse;
	InitDesc();
	IsStack = FALSE;
	Palette = new CPalette;
	DetectVideoModes();
}

CGraph::~CGraph()
{
	delete Palette;
}

void CGraph::DetectVideoModes()
{
	TModeInfo ModeInfo;
	TSVGAInfo SVGAInfo;
	unsigned short *ModeList;
	int Index, Count;

	if (GetSVGAInfo(&SVGAInfo))
		return;
	GraphInfo.Name = SVGAInfo.name;
	GraphInfo.VersionMaj = SVGAInfo.vermaj;
	GraphInfo.VersionMin = SVGAInfo.vermin;
	GraphInfo.Memory = SVGAInfo.memory;

	for (ModeList = SVGAInfo.modes; *ModeList != 0xffff; ++ModeList) {
		if (!GetModeInfo(*ModeList,&ModeInfo) &&
			 (ModeInfo.mode_attr & 1) != 0 && ModeInfo.bits_pixel == 8)
			if (ModeInfo.width == 640 && ModeInfo.height == 480)
				AddVideoMode(mode640x480,*ModeList,ModeInfo);
			else if (ModeInfo.width == 800 && ModeInfo.height == 600)
				AddVideoMode(mode800x600,*ModeList,ModeInfo);
			else if (ModeInfo.width == 1024 && ModeInfo.height == 768)
				AddVideoMode(mode1024x768,*ModeList,ModeInfo);
			else if (ModeInfo.width == 1280 && ModeInfo.height == 1024)
				AddVideoMode(mode1280x1024,*ModeList,ModeInfo);
			else if (ModeInfo.width == 1600 && ModeInfo.height == 1200)
				AddVideoMode(mode1600x1200,*ModeList,ModeInfo);
	}
	for (Count = 0, Index = 1; Index < 6; ++Index)
		if (VesaModeList[Index].VesaMode != -1)
			++Count;
		Count = 0;
	if (!Count) {
		// either VESA modes our not supported, or ModeList is empty
		// try the default VesaModes
		if (!GetModeInfo(0x0101,&ModeInfo) && (ModeInfo.mode_attr & 1) != 0)
			AddVideoMode(mode640x480,0x0101,ModeInfo);
		if (!GetModeInfo(0x0103,&ModeInfo) && (ModeInfo.mode_attr & 1) != 0)
			AddVideoMode(mode800x600,0x0103,ModeInfo);
		if (!GetModeInfo(0x0105,&ModeInfo) && (ModeInfo.mode_attr & 1) != 0)
			AddVideoMode(mode1024x768,0x0105,ModeInfo);
		if (!GetModeInfo(0x0107,&ModeInfo) && (ModeInfo.mode_attr & 1) != 0)
			AddVideoMode(mode1280x1024,0x0107,ModeInfo);
	}
}

void CGraph::AddVideoMode(TGraphMode GraphMode, int VesaMode, TModeInfo &ModeInfo)
{
	VesaModeList[GraphMode].VesaMode = VesaMode;
	VesaModeList[GraphMode].LFBSupport = ModeInfo.mode_attr & 0x80;
	VesaModeList[GraphMode].LinearBuffer = ModeInfo.linear_buffer;
	VesaModeList[GraphMode].Granularity = ModeInfo.granularity;
	VesaModeList[GraphMode].SwitchAddr = ModeInfo.switch_addr;
}

int CGraph::SetMode(TGraphMode Mode)
{
	int VesaMode = VesaModeList[Mode].VesaMode;

	if (VesaMode == -1)
		return -1;
	if (SwitchTo(VesaMode))
		return -1;
	if (VesaMode != 3) {
		GRWidth = VesaModeList[Mode].Width;
		GRHeight = VesaModeList[Mode].Height;
		GRThisBank = 0;
		LFBSupport = VesaModeList[Mode].LFBSupport;
		AddrLinearFB = VesaModeList[Mode].LinearBuffer;
		if (LFBSupport)
			if (!ConfirmLFB(AddrLinearFB))
				LFBSupport = FALSE;
		CreateBankswitch(VesaModeList[Mode].Granularity,VesaModeList[Mode].SwitchAddr);

		GraphInfo.Width = GRWidth;
		GraphInfo.Height = GRHeight;

		ModeName = VesaModeList[Mode].ModeName;
		InitViewport(GRWidth,GRHeight);
		Palette->Install();
	}
	this->Mode = Mode;
	return 0;
}

TGraphMode CGraph::GetMode()
{
	return Mode;
}

TGraphMode CGraph::GetModeFromIndex(int ModeIndex)
{
	int Index;

	if (ModeIndex < GetModeCount()) {
		for (Index = 0; Index != ModeIndex; )
			if (VesaModeList[Index + 1].VesaMode != -1)
				++Index;
		return (TGraphMode)(Index + 1);
	}
	return mode640x480;
}

TGraphMode CGraph::GetModeFromMetrics(int Width, int Height)
{
	if (Width == 640 && Height == 480)
		return mode640x480;
	if (Width == 800 && Height == 600)
		return mode800x600;
	if (Width == 1024 && Height == 768)
		return mode1024x768;
	if (Width == 1280 && Height == 1024)
		return mode1280x1024;
	if (Width == 1600 && Height == 1200)
		return mode1600x1200;
	return mode640x480;
}

int CGraph::GetModeIndex(TGraphMode GraphMode)
{
	int Index, ModeIndex;

	for (Index = 1, ModeIndex = 0; Index < GraphMode; ++Index)
		if (VesaModeList[Index].VesaMode != -1)
			++ModeIndex;
	return ModeIndex;
}

int CGraph::GetModeCount()
{
	int Index, Count = 0;

	for (Index = mode640x480; Index <= mode1600x1200; ++Index)
		if (VesaModeList[Index].VesaMode != -1)
			 ++Count;
	return Count;
}

void CGraph::VBE2(int Enabled)
{
	if (!Enabled)
		InitVBE2(0);
	else
		if (LFBSupport)
			InitVBE2(AddrLinearFB);
}

int CGraph::LinearSupport()
{
	return LFBSupport;
}

const char *CGraph::GetModeName()
{
	return ModeName;
}

const char *CGraph::GetModeName(int Mode)
{
	return Mode >= 0 && Mode <= mode1600x1200 ? VesaModeList[Mode].ModeName : "";
}


void CGraph::GetModeMetrics(int &Width, int &Height)
{
	Width = GraphInfo.Width;
	Height = GraphInfo.Height;
}

void CGraph::GetGraphInfo(TGraphInfo &GraphInfo)
{
	memcpy(&GraphInfo,&this->GraphInfo,sizeof (TGraphInfo));
}

int CGraph::ModeSupported(TGraphMode Mode)
{
	return VesaModeList[Mode].VesaMode != -1;
}

void CGraph::FlushQueue()
{
	PGraphQueue Temp;
	PGraphQueue Free;

	DoPMGraphics();
	Temp = GraphQueue.Next;
	while (Temp) {
		Free = Temp;
		Temp = Temp->Next;
		delete Free->Params;
		delete Free;
	}
	GraphQueue.Next = NULL;
	GraphQueue.PhysNext = 0;
	Tail = &GraphQueue;
}

void CGraph::CreateBankswitch(int Granularity, unsigned long SwitchAddr)
// pure memory manipulation...
{
	int shl = 0;
	int bmul = 64 / Granularity;
	char far *switchproc = (char far *)BankSwitch;

	for (; bmul >>= 1; ++shl);
	if (shl) {
		*switchproc++ = 0x52; // push dx
		*((unsigned short far *)switchproc)++ = 0xe2c1; // shl dx,...
		*switchproc++ = shl; // shl dx,_shl
	}
	*switchproc++ = 0x9a; // call ....:....
	*((unsigned long far *)switchproc)++ = SwitchAddr;
	if (shl)
		*switchproc++ = 0x5a; // pop dx
	*switchproc = 0xc3; // ret
}

void CGraph::QFlushArea(long Left, long Top, long Width, long Height)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height);
	QueueAction(VPFlushArea,ParamList);
}

void CGraph::QUltraFlush()
{
	QueueAction(VPUltraFlush,CreateParamList());
}

void CGraph::QStoreBuffer(long Left, long Top, long Width, long Height)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height);
	QueueAction(VPStoreBuffer,ParamList);
}

void CGraph::QRestoreBuffer(long Left, long Top, long Width, long Height)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height);
	QueueAction(VPRestoreBuffer,ParamList);
}

void CGraph::QueueAction(void *Func,long far *ParamList)
{
	PGraphQueue NewItem;

	NewItem = new TGraphQueue;
	if (!IsStack) {
		Tail->Next = NewItem;
		Tail->PhysNext = PhysAddr(NewItem);
		Tail = Tail->Next;
		NewItem->Next = NULL;
		NewItem->PhysNext = 0;
	}
	else {
		NewItem->Next = GraphQueue.Next;
		NewItem->PhysNext = PhysAddr(GraphQueue.Next);
		GraphQueue.Next = NewItem;
		GraphQueue.PhysNext = PhysAddr(NewItem);
		if (Tail == &GraphQueue)
			Tail = NewItem;
	}
	NewItem->Params = ParamList;
	NewItem->PhysParams = PhysAddr(ParamList);
	NewItem->Func = FP_OFS(Func);
}

long far *CGraph::CreateParamList(long P1, long P2, long P3, long P4, long P5)
{
	long far *ParamList;

	ParamList = new long[5];
	ParamList[4] = P1;
	ParamList[3] = P2;
	ParamList[2] = P3;
	ParamList[1] = P4;
	ParamList[0] = P5;
	return ParamList;
}

void CGraph::QSetViewportOrigin(long &Left, long &Top)
{
	long far *ParamList;

	ParamList = CreateParamList(PhysAddr(&Left),PhysAddr(&Top));
	QueueAction(SetViewportOrigin,ParamList);
}
void CGraph::QGetViewportOrigin(long &Left, long &Top)
{
	long far *ParamList;

	ParamList = CreateParamList(PhysAddr(&Left),PhysAddr(&Top));
	QueueAction(GetViewportOrigin,ParamList);
}

void CGraph::QSetClippingRegion(long &Left, long &Top, long &Width, long &Height)
{
	long far *ParamList;

	ParamList = CreateParamList(PhysAddr(&Left),PhysAddr(&Top),PhysAddr(&Width),PhysAddr(&Height));
	QueueAction(SetClippingRegion,ParamList);
}

void CGraph::QGetClippingRegion(long &Left, long &Top, long &Width, long &Height)
{
	long far *ParamList;

	ParamList = CreateParamList(PhysAddr(&Left),PhysAddr(&Top),PhysAddr(&Width),PhysAddr(&Height));
	QueueAction(GetClippingRegion,ParamList);
}

void CGraph::QPutPixel(long Left, long Top, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Color);
	QueueAction(VPPutPixel,ParamList);
}

void CGraph::QHLine(long Left, long Top, long Width, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Color);
	QueueAction(VPHLine,ParamList);
}

void CGraph::QVLine(long Left, long Top, long Height, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Height,Color);
	QueueAction(VPVLine,ParamList);
}

void CGraph::QRectangle(long Left, long Top, long Width, long Height, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height,Color);
	QueueAction(VPRectangle,ParamList);
}

void CGraph::QBar(long Left, long Top, long Width, long Height, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height,Color);
	QueueAction(VPBar,ParamList);
}

void CGraph::QPutImage(long Left, long Top, long Width, long Height, const void far *Image)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height,PhysAddr(Image));
	QueueAction(VPPutImage,ParamList);
}

void CGraph::QGetImage(long Left, long Top, long Width, long Height, const void far *Image)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,Width,Height,PhysAddr(Image));
	QueueAction(VPGetImage,ParamList);
}

void CGraph::QLine(long X1, long Y1, long X2, long Y2, long Color)
{
	long far *ParamList;

	ParamList = CreateParamList(X1,Y1,X2,Y2,Color);
	QueueAction(VPLine,ParamList);
}

void CGraph::SetIsStack(int Status)
{
	IsStack = Status;
}

void CGraph::WaitRetrace()
{
	::WaitRetrace();
}

void CGraph::SetCursorXY(int X, int Y)
{
	if (CursorVisible) {
		QFlushArea(X,Y,CURSOR_WIDTH,CURSOR_HEIGHT);
		QFlushArea(CursorX,CursorY,CURSOR_WIDTH,CURSOR_HEIGHT);
	}
	CursorX = X;
	CursorY = Y;
	if (CursorVisible)
		FlushQueue();
}

void CGraph::GetCursorXY(int &X, int &Y)
{
	X = CursorX;
	Y = CursorY;
}

void CGraph::ShowCursor()
{
	if (Mouse->Available() != -1) {
		CursorVisible = TRUE;
		QFlushArea(CursorX,CursorY,CURSOR_WIDTH,CURSOR_HEIGHT);
		FlushQueue();
	}
}

void CGraph::HideCursor()
{
	CursorVisible = FALSE;
	QFlushArea(CursorX,CursorY,CURSOR_WIDTH,CURSOR_HEIGHT);
	FlushQueue();
}

void CGraph::SetFont(void *FontFile)
{
	int *pFontFile;

	pFontFile = (int *)FontFile;

	//FontHeader = new TFontEntry[256];

	DataSize = *pFontFile++;

	TextHeight = *pFontFile++;

//	memcpy(FontHeader,pFontFile,sizeof (TFontEntry) * 256);
	FontHeader = (TFontEntry *)pFontFile;
	pFontFile += sizeof (TFontEntry) * 128;
	PhysFontHeader = PhysAddr(FontHeader);
	PhysFontData = PhysAddr(pFontFile);
//	MapFontData();
}

int CGraph::GetTextWidth(const char *Str, int Style)
{
	int Length;
	const unsigned char *pStr;

	pStr = (const unsigned char *)Str;

	for (Length = 0; *pStr; Length += FontHeader[*pStr++].Width);
	if (Style == STYLE_BOLD)
		++Length;
	return Length;
}

int CGraph::GetTextHeight()
{
	return TextHeight;
}

void CGraph::QTextOut(int Left, int Top, const char *Str, int Style, int Color)
{
	long far *ParamList;

	ParamList = CreateParamList(Left,Top,PhysAddr(Str),Color);
	QueueAction(VPTextOut,ParamList);
	if (Style == STYLE_BOLD) {
		ParamList = CreateParamList(Left + 1,Top,PhysAddr(Str),Color);
		QueueAction(VPTextOut,ParamList);
	}
}

