/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <edit.h>
#include <graph.h>
#include <string.h>
#include <key.h>

extern void printf(const char *fmt, ...);

CEdit::CEdit(const char *Text, int MaxLength, int Masked, int ReadOnly,
	int Left, int Top, int Width, int Visible, void *HandlerClass):
	CWindow(Left,Top,Width,21,Visible,FALSE,TRUE,HandlerClass)
{
	Caption = Text;
	this->MaxLength = MaxLength;
	this->Masked = Masked;
	this->ReadOnly = ReadOnly;
	Cursor = 0;
}


CEdit::~CEdit()
{
}

void CEdit::SetMaxLength(int MaxLength)
{
	this->MaxLength = MaxLength;
}

void CEdit::SetMasked(int Masked)
{
	this->Masked = Masked;
}

void CEdit::SetReadOnly(int ReadOnly)
{
	this->ReadOnly = ReadOnly;
}

void CEdit::Focus()
{
	Cursor = Caption.Length();
	CWindow::Focus();
}

void CEdit::Draw(long,long,long,long)
{
	int TextTop, TextLeft;
	int TextWidth;
	const char *DisplayStr;

	TextTop = 21 - Graph->GetTextHeight() >> 1;
	Graph->QBar(2,2,Width - 2,18,21);

	if (!Masked)
		DisplayStr = Caption;
	else {
		Mask.CharFill('*',Caption.Length());
		DisplayStr = Mask;
	}
	if (!GotFocus) {
		TextWidth = Graph->GetTextWidth(DisplayStr,STYLE_REGULAR);
		if (TextWidth > Width - 5)
			TextLeft = Width - TextWidth - 5;
		else
			TextLeft = 3;
	}
	else {
		CString Str, Str1, Str2;

		Str = DisplayStr;
		Str.Split(Cursor,Str1,Str2);
		TextWidth = Graph->GetTextWidth(Str1,STYLE_REGULAR);
		if (TextWidth > Width - 5)
			TextLeft = Width - TextWidth - 5;
		else
			TextLeft = 3;
	}
	Graph->QTextOut(TextLeft,TextTop,(const char *)DisplayStr,STYLE_REGULAR,17);
	if (GotFocus)
		Graph->QVLine(TextLeft + TextWidth,3,15,17);

	Graph->QHLine(0,0,Width - 1,17);
	Graph->QVLine(0,1,Height - 2,17);

	Graph->QHLine(1,1,Width - 3,18);
	Graph->QVLine(1,2,Height - 4,18);

	Graph->QHLine(0,Height - 1,Width,21);
	Graph->QVLine(Width - 1,0,Height - 1,21);

	Graph->QHLine(1,Height - 2,Width - 2,19);
	Graph->QVLine(Width - 2,1,Height - 3,19);
}

void CEdit::KeyPress(int Key)
{
	if (ReadOnly)
		return;
	if (WndOnKeyPress && HandlerClass)
		WndOnKeyPress(HandlerClass,Key);
	switch (Key) {
		case KEY_NONE:
			return;
		case KEY_BACKSPACE:
			if (Cursor) {
				Caption.DeleteChar(Cursor - 1);
				--Cursor;
				Refresh();
			}
			break;
		case KEY_LEFT:
			if (Cursor) {
				--Cursor;
				Refresh();
			}
			break;
		case KEY_RIGHT:
			if (Cursor < Caption.Length()) {
				++Cursor;
				Refresh();
			}
			break;
		case KEY_DELETE:
			if (Cursor != Caption.Length()) {
				Caption.DeleteChar(Cursor);
				Refresh();
			}
			break;
		case KEY_HOME:
			Cursor = 0;
			Refresh();
			break;
		case KEY_END:
			Cursor = Caption.Length();
			Refresh();
			break;
		case KEY_PAGEUP:
		case KEY_CTRLLEFT:
			Cursor -= 4;
			if (Cursor < 0)
				Cursor = 0;
			Refresh();
			break;
		case KEY_PAGEDOWN:
		case KEY_CTRLRIGHT:
			Cursor += 4;
			if (Cursor > Caption.Length())
				Cursor = Caption.Length();
			Refresh();
			break;
		case KEY_ESCAPE:
			Caption.Clear();
			Cursor = 0;
			Refresh();
			break;
		default:
			if (!MaxLength || Caption.Length() < MaxLength)
				if ((Key &= 0x00ff) != 0) {
					Caption.InsertChar(Cursor,Key & 0x00ff);
					++Cursor;
					Refresh();
				}
			break;
	}
}

void CEdit::SetText(const char *Text)
{
	SetCaption(Text);
	Cursor = Caption.Length();
}

void CEdit::GetText(char *Text)
{
	strcpy(Text,Caption);
}

int CEdit::MouseDown(int MouseX, int MouseY)
{
	return CWindow::MouseDown(MouseX, MouseY);

}
