/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <listbox.h>
#include <graph.h>
#include <key.h>
#include <scroll.h>
#include <text.h>

CListBox::CListBox(int Columns, int ShowHeader,
	int Left, int Top, int Width, int Height, int Visible, void *HandlerClass):
	CWindow(Left,Top,Width,Height,Visible,FALSE,TRUE,HandlerClass)
{
	int DrawHeight;

	ScrollBar = NULL;
	ListBoxSelect = NULL;
	Count = 0;
	ItemIndex = -1;
	this->Columns = Columns;
	this->ShowHeader = ShowHeader;
	DrawStart = 0;
	DrawHeight = Height - 4;
	if (ShowHeader)
		DrawHeight -= 20;
	DrawCount = DrawHeight / (Graph->GetTextHeight() + 1);
	ColumnHeaders = new CStringList(Columns);
	ColumnWidth = new int[Columns];
	while (Columns)
		ColumnWidth[--Columns] = 0;
	Items = new TStrListNode;
	Items->Next = NULL;
	Items->Prev = NULL;
}

CListBox::~CListBox()
{
	Clear();
	delete Items;
	delete ColumnWidth;
	delete ColumnHeaders;
}

void CListBox::SetColumn(int Index, int Width, const char *HeaderName)
{
	if (Index >= Columns)
		return;
	ColumnWidth[Index] = Width;
	ColumnHeaders->Insert(Index,HeaderName);
}

void CListBox::SetShowHeader(int ShowHeader)
{
	this->ShowHeader = ShowHeader;
	Refresh();
}

int CListBox::AddRow()
{
	InsertRow(Count);
	return Count - 1;
}

void CListBox::InsertRow(int Index)
{
	TStrListNode *Entry;
	TStrListNode *NewRow;

	if (Index > Count)
		Index = Count;
	Entry = GetEntry(Index - 1);
	NewRow = new TStrListNode;
	NewRow->StringList = new CStringList(Columns);
	NewRow->Next = Entry->Next;
	NewRow->Prev = Entry;
	if (Entry->Next)
		Entry->Next->Prev = NewRow;
	Entry->Next = NewRow;
	++Count;
	UpdateScrollBar();
}

void CListBox::DeleteRow(int Index)
{
	TStrListNode *Entry;

	if (Index >= Count)
		return;
	Entry = GetEntry(Index);
	Entry->Prev->Next = Entry->Next;
	if (Entry->Next)
		Entry->Next->Prev = Entry->Prev;
	delete Entry->StringList;
	delete Entry;
	--Count;
	UpdateScrollBar();
}

void CListBox::AddItem(int Row, int Column, const char *String)
{
	TStrListNode *Entry;

	if (Row >= Count)
		return;
	Entry = GetEntry(Row);
	Entry->StringList->Insert(Column,String);
}

CString CListBox::GetItem(int Row, int Column)
{
	TStrListNode *Entry;

	if (Row >= Count)
		return "";
	Entry = GetEntry(Row);
	return Entry->StringList->Get(Column);
}

void CListBox::Clear()
{
	TStrListNode *Entry;
	TStrListNode *Temp;

	for (Entry = Items->Next; Entry;) {
		Temp = Entry;
		Entry = Entry->Next;
		delete Temp->StringList;
		delete Temp;
	}
	Items->Next = NULL;
	Count = 0;
	ItemIndex = -1;
	UpdateScrollBar();
}

void CListBox::SetItemIndex(int Index)
{
	if (Index >= Count)
		Index = -1;
	if (ItemIndex != Index) {
		ItemIndex = Index;
		if (ItemIndex < 0)
			DrawStart = 0;
		else
			if (ItemIndex < DrawStart)
				DrawStart = ItemIndex;
			else
				while (ItemIndex >= DrawStart + DrawCount)
					++DrawStart;
		Refresh();
		if (ScrollBar)
			ScrollBar->SetValue(DrawStart);
		if (ListBoxSelect && HandlerClass)
			ListBoxSelect(HandlerClass,ItemIndex);
	}
}

int CListBox::GetItemIndex()
{
	return ItemIndex;
}

int CListBox::GetCount()
{
	return Count;
}

void CListBox::SetDrawStart(int Index)
{
	if (DrawStart != Index) {
		this->DrawStart = Index;
		Refresh();
		if (ScrollBar)
			ScrollBar->SetValue(DrawStart);
	}
}

int CListBox::GetDrawCount()
{
	return DrawCount;
}

TStrListNode *CListBox::GetEntry(int Index)
{
	TStrListNode *Entry;

	for (Entry = Items; Index >= 0; --Index)
		Entry = Entry->Next;
	return Entry;
}

void CListBox::Draw(long, long, long, long)
{
	Graph->QBar(2,2,Width - 4,Height - 4,21);
	DrawItems();
	DrawBody();
	if (ShowHeader)
		DrawHeader();
}

void CListBox::DrawBody()
{
	Graph->QHLine(0,0,Width - 1,18);
	Graph->QVLine(0,1,Height - 2,18);
	Graph->QHLine(1,1,Width - 3,17);
	Graph->QVLine(1,2,Height - 4,17);
	Graph->QHLine(0,Height - 1,Width,21);
	Graph->QVLine(Width - 1,0,Height - 1,21);
	Graph->QHLine(1,Height - 1,Width - 2,20);
	Graph->QVLine(Width - 2,1,Height - 3,20);
}

void CListBox::DrawHeader()
{
	int Index;
	int Left, Width;
	int TextTop;

	TextTop = 20 - Graph->GetTextHeight();
	Left = 2;
	for (Index = 0; Index < Columns; ++Index) {
		Width = ColumnWidth[Index];
		Graph->QBar(Left + 2,4,Width - 4,16,19);
		Graph->QTextOut(Left + 3,TextTop,ColumnHeaders->Get(Index),STYLE_REGULAR,17);
		Graph->QHLine(Left,2,Width - 1,21);
		Graph->QVLine(Left,3,18,21);
		Graph->QHLine(Left + 1,3,Width - 3,20);
		Graph->QVLine(Left + 1,4,16,20);
		Graph->QHLine(Left + 1,20,Width - 2,18);
		Graph->QHLine(Left,21,Width,17);
		Left += Width;
		Graph->QVLine(Left - 2,3,17,18);
		Graph->QVLine(Left - 1,2,19,17);
	}
}

void CListBox::DrawItems()
{
	int Col, Row;
	long OldLeft, OldTop, OldWidth, OldHeight;
	long ClipLeft, ClipTop, ClipWidth, ClipHeight;
	int AbsLeft, AbsTop;
	TStrListNode *Entry;

	int TextLeft, TextTop, ColWidth;
	int TextStart;
	int DrawEnd;
	int RowHeight;

	if (!Count)
		return;
	RowHeight = Graph->GetTextHeight() + 1;
	TextStart = ShowHeader ? 22 : 2;
	GetAbsPosition(AbsLeft, AbsTop);
	ClipTop = AbsTop + TextStart;
	ClipHeight = Height - 2 - TextStart;

	DrawEnd = DrawStart + DrawCount;
	TextLeft = 5;
	for (Col = 0; Col < Columns; ++Col) {
		ColWidth = ColumnWidth[Col];

//		Graph->QGetClippingRegion(OldLeft, OldTop, OldWidth, OldHeight);
		ClipLeft = AbsLeft + TextLeft - 3;
		ClipWidth = ColWidth;
//		Graph->QSetClippingRegion(ClipLeft,ClipTop,ClipWidth,ClipHeight);
		/***/

		TextTop = TextStart;
		Entry = GetEntry(DrawStart);
		for (Row = DrawStart; Entry && Row <= DrawEnd; ++Row) {
			if (Row != ItemIndex)
				Graph->QTextOut(TextLeft,TextTop,Entry->StringList->Get(Col),STYLE_REGULAR,17);
			else {
				if (GotFocus)
					Graph->QBar(TextLeft - 3,TextTop,ColWidth,RowHeight,17);
				else
					Graph->QBar(TextLeft - 3,TextTop,ColWidth,RowHeight,18);
				Graph->QTextOut(TextLeft,TextTop,Entry->StringList->Get(Col),STYLE_REGULAR,21);
			}

			Entry = Entry->Next;
			TextTop += RowHeight;
		}
		/***/
		TextLeft += ColWidth;
//		Graph->QSetClippingRegion(OldLeft, OldTop, OldWidth, OldHeight);
//		Graph->FlushQueue();
	}
}

void CListBox::KeyPress(int Key)
{
	int Index;

	Index = ItemIndex;
	switch (Key) {
		case KEY_UP:
			if (Index > 0) {
				--Index;
				SetItemIndex(Index);
			}
			break;
		case KEY_DOWN:
			if (Index < Count - 1) {
				++Index;
				SetItemIndex(Index);
			}
			break;
		case KEY_HOME:
			SetItemIndex(0);
			break;
		case KEY_END:
			SetItemIndex(Count - 1);
			break;
		case KEY_PAGEUP:
			Index -= DrawCount - 1;
			if (Index < 0)
				Index = 0;
			SetItemIndex(Index);
			break;
		case KEY_PAGEDOWN:
			Index += DrawCount - 1;
			if (Index >= Count)
				Index = Count - 1;
			SetItemIndex(Index);
			break;
		default:
			break;
	}
}

int CListBox::MouseDown(int MouseX, int MouseY)
{
	int Status;
	int Index;

	Status = CWindow::MouseDown(MouseX,MouseY);
	if (Status != -1 && Enabled) {
		MouseY -= Top + 1;
		if (ShowHeader)
			MouseY -= 20;
		Index = MouseY / (Graph->GetTextHeight() + 1) + DrawStart;
		if (Index < Count && MouseY > 0)
			SetItemIndex(Index);
	}
	return Status;
}

void CListBox::SetScrollBar(CScrollBar *ScrollBar)
{
	this->ScrollBar = ScrollBar;
	ScrollBar->OnChange((TScrollChange)ScrollBarChange);
	UpdateScrollBar();
}

void CListBox::ScrollBarChange(CListBox *ListBox, int Value)
{
	ListBox->SetDrawStart(Value);
}

void CListBox::UpdateScrollBar()
{
	int ScrollCount;

	if (!ScrollBar)
		return;
	ScrollCount = Count - DrawCount;
	if (ScrollCount <= 0)
		ScrollBar->Disable();
	else {
//		ScrollBar->Parent->SetVisible(FALSE);
		ScrollBar->SetMin(0);
		ScrollBar->SetMax(ScrollCount);
		ScrollBar->SetValue(DrawStart);
		ScrollBar->Enable();

//		ScrollBar->Parent->SetVisible(TRUE);
//		ScrollBar->Refresh();
	}
}

void CListBox::OnSelect(TListBoxSelect ListBoxSelect)
{
	this->ListBoxSelect = ListBoxSelect;
}

void CListBox::FontChanged()
{
	int DrawHeight;

	DrawHeight = Height - 4;
	if (ShowHeader)
		DrawHeight -= 20;
	DrawCount = DrawHeight / (Graph->GetTextHeight() + 1);
	UpdateScrollBar();
}
