/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <screen.h>
#include <graph.h>
#include <backgnd.h>
#include <window.h>
#include <cursor.h>
#include <mouse.h>
#include <text.h>

/*
 * The screen consists of four layers:
 * layer 1: background; drawn by BeforeFix()
 * layer 2: all controls (forms)
 * layer 3: 'foreground'; drawn by AfterFix() (empty at the moment)
 * layer 4: mouse cursor
 */

CScreen::CScreen()
{
	IgnoreSU = FALSE;
	Background = new CBackground;
	WindowList = new CWindowList();
	WindowList->SetPosition(0,0);
	WindowList->SetHandler(this);
	UseWallpaper = FALSE;
	MouseX = -1;
	MouseY = -1;
	MouseDown = 0;
}

CScreen::~CScreen()
{
	delete Background;
	delete WindowList;
}

int CScreen::ShowSplashScreen(TGraphMode  Mode, int LinearFB)
{
	int Width, Height;

	while (Mode != modeText && Graph->SetMode(Mode) == -1)
		--(int)Mode;
	if (Mode == modeText)
		return -1;
	Graph->SetMode(Mode);
	if (Graph->LinearSupport())
		Graph->VBE2(LinearFB);
	Graph->GetModeMetrics(Width,Height);
	Background->SplashScreen(Width,Height);
	Graph->QUltraFlush();
	Graph->FlushQueue();
	return 0;
}

void CScreen::BeforeFix(int Left, int Top, int Width, int Height)
{
	if (!AdjustArea(Left,Top,Width,Height))
		return;
	Background->FixDamage(Left,Top,Width,Height);
}

void CScreen::AfterFix(int, int, int, int)
{
}

void CScreen::AddWindow(CWindow *Wnd)
{
	WindowList->Add(Wnd);
}

void CScreen::RemoveWindow(CWindow *Wnd)
{
	WindowList->Remove(Wnd);
}

int CScreen::SetGraphicsMode(TGraphMode  Mode)
{
	if (Graph->SetMode(Mode) == -1 || Mode == modeText)
		return -1;
	InitScreen();
	return 0;
}

int CScreen::TestGraphicsMode(TGraphMode  Mode)
{
	int Width, Height;

	if (Graph->SetMode(Mode) == -1)
		return -1;
	Graph->GetModeMetrics(Width,Height);
	Background->DrawTestScreen();
	Graph->QUltraFlush();
	Graph->FlushQueue();
	return 0;
}

void CScreen::GetScreenSize(int &Width, int &Height)
{
	Width = ScrnWidth;
	Height = ScrnHeight;
}

int CScreen::AdjustArea(int &Left, int &Top, int &Width, int &Height)
{
	long Bottom, Right;

	if (Top >= ScrnHeight || Left >= ScrnWidth)
		return FALSE;
	if (Top < 0) {
		Height += Top;
		Top = 0;
	}
	Bottom = Top + Height - 1;
	if (Bottom >= ScrnBottom)
		Height -= Bottom - ScrnBottom;
	if (Left < 0) {
		Width += Left;
		Left = 0;
	}
	Right = Left + Width - 1;
	if (Right > ScrnRight)
		Width -= Right - ScrnRight;
	if (Width > 0 && Height > 0)
		return TRUE;
	return FALSE;
}

void CScreen::SetFocusWnd(CWindow *Wnd)
{
	WindowList->SetFocusWnd(Wnd,TRUE);
}

CWindow *CScreen::GetFocusWnd()
{
	return WindowList->GetFocusWnd();
}

void CScreen::Refresh()
{
	if (!IgnoreSU) {
		WindowList->FixDamage(0,0,ScrnWidth,ScrnHeight);
		Graph->QUltraFlush();
		Graph->FlushQueue();
	}
}

void CScreen::MouseStatus(int Left, int Top, int ButtonDown)
{
	if (ButtonDown != MouseDown) {
		MouseDown = ButtonDown;
		if (MouseDown)
			WindowList->MouseDown(Left,Top);
		else
			WindowList->MouseUp();
	}
	if (Left != MouseX || Top != MouseY) {
		MouseX = Left;
		MouseY = Top;
		WindowList->MouseMove(Left,Top);
	}
}

void CScreen::KeyPress(int Key)
{
	CWindow *Wnd;

	Wnd = WindowList->GetFocusWnd();
	if (Wnd && Wnd->Visible)
		Wnd->KeyPress(Key);
}

void CScreen::SetUseWallpaper(int UseWallpaper, int ScrnRefresh)
{
	if (UseWallpaper != this->UseWallpaper) {
		this->UseWallpaper = UseWallpaper;
		if (!IgnoreSU && ScrnRefresh) {
			Background->NewGraphicsMode(ScrnWidth,ScrnHeight,UseWallpaper);
			Refresh();
		}
	}
}

void CScreen::SetWallpaper(int Width, int Height, const unsigned char *Wallpaper)
{
	Background->SetWallpaper(Width,Height,Wallpaper);
}

void CScreen::SetSplashLogo(int Width, int Height, const unsigned char *LogoBitmap)
{
	Background->SetSplashLogo(Width,Height,LogoBitmap);
}

void CScreen::SetIgnoreSU(int IgnoreSU)
{
	this->IgnoreSU = IgnoreSU;
}

void CScreen::InitScreen()
{
	Graph->GetModeMetrics(ScrnWidth,ScrnHeight);
	ScrnRight = ScrnWidth - 1;
	ScrnBottom = ScrnHeight - 1;
	Background->NewGraphicsMode(ScrnWidth,ScrnHeight,UseWallpaper);

	WindowList->SetMetrics(ScrnWidth,ScrnHeight);
	WindowList->FixDamage(0,0,ScrnWidth,ScrnHeight);
	Graph->QUltraFlush();
	Graph->FlushQueue();
}
