/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <wndlist.h>
#include <graph.h>
#include <screen.h>
#include <window.h>
#include <form.h>

/*
 * NOTES:
 *
 * never delete a focused window
 * ThisNode undefined after delete
 */

/*
 * WidowList == dummy. Index == -1
 * WindowList---Item0---Item1---...
 */

CWindowList::CWindowList()
{
	WindowList = new TWndNode;
	WindowList->Next = WindowList->Prev = NULL;
	LastNode = WindowList;
	WndCount = 0;
	FocusEntry = NULL;
	hScreen = NULL;
	hForm = NULL;
	Visible = TRUE;
	LastFocus = NULL;
}

CWindowList::~CWindowList()
{
	PWndNode Temp;

	while (WindowList) {
		Temp = WindowList;
		WindowList = WindowList->Next;
		delete Temp;
	}
}

void CWindowList::SetVisible(int Visible)
{
	this->Visible = Visible;
}

int CWindowList::GetVisible()
{
	return Visible;
}

void CWindowList::Add(CWindow *Wnd)
{
	PWndNode NewNode;

	NewNode = new TWndNode;
	NewNode->Wnd = Wnd;
	NewNode->Next = NULL;
	NewNode->Prev = LastNode;
	ThisNode = NULL;
	LastNode = LastNode->Next = NewNode;
	++WndCount;
	Wnd->SetParent(this);
}

void CWindowList::Insert(int Index, CWindow *Wnd)
{
	PWndNode NewNode, PrevNode;

	if (Index >= WndCount)
		Add(Wnd);
	else {
		for (PrevNode = WindowList; Index--; PrevNode = PrevNode->Next);
		NewNode = new TWndNode;
		NewNode->Wnd = Wnd;
		NewNode->Next = PrevNode->Next;
		NewNode->Prev = PrevNode;
		PrevNode->Next = NewNode;
		NewNode->Next->Prev = NewNode;
		++WndCount;
		Wnd->SetParent(this);
	}

}

void CWindowList::Remove(CWindow *Wnd)
{
	PWndNode DelNode;

	for (DelNode = WindowList->Next; DelNode && DelNode->Wnd != Wnd; DelNode = DelNode->Next);
	if (!DelNode)
		return;
	DelNode->Prev->Next = DelNode->Next;
	if (DelNode->Next)
		DelNode->Next->Prev = DelNode->Prev;
	else
		LastNode = DelNode->Prev;

	delete DelNode;
	--WndCount;
}

CWindow *CWindowList::Items(int Index)
{
	PWndNode INode;

	if (Index >= WndCount)
		return NULL;
	for (INode = WindowList->Next; INode--; INode = INode->Next);
	return INode->Wnd;
}

void CWindowList::Delete(int Index)
{
	PWndNode DelNode;

	if (Index >= WndCount)
		return;
	for (DelNode = WindowList->Next; Index--; DelNode = DelNode->Next);
	DelNode->Prev->Next = DelNode->Next;
	DelNode->Next->Prev = DelNode->Prev;
	delete DelNode;
	--WndCount;
}

int CWindowList::Count()
{
	return WndCount;
}

void CWindowList::GotoFirst()
{
	ThisNode = WindowList->Next;
}

void CWindowList::GotoLast()
{
	if (LastNode == WindowList)
		ThisNode = NULL;
	else
		ThisNode = LastNode;
}

CWindow *CWindowList::GetThis()
{
	if (ThisNode)
		return ThisNode->Wnd;
	return NULL;
}

CWindow *CWindowList::GetNext()
{
	if (ThisNode && (ThisNode = ThisNode->Next) != NULL)
		return ThisNode->Wnd;
	return NULL;
}

CWindow *CWindowList::GetPrev()
{
	if (ThisNode) {
		ThisNode = ThisNode->Prev;
		if (ThisNode == WindowList)
			return NULL;
		return ThisNode->Wnd;
	}
	return NULL;
}

void CWindowList::SetPosition(int Left, int Top)
{
	this->Left = Left;
	this->Top = Top;
}

void CWindowList::GetPosition(int &Left, int &Top)
{
	Left = this->Left;
	Top = this->Top;
}

void CWindowList::SetMetrics(int Width, int Height)
{
	this->Width = Width;
	this->Height = Height;
}

void CWindowList::GetMetrics(int &Width, int &Height)
{
	Width = this->Width;
	Height = this->Height;
}

void CWindowList::SetFocusWnd(CWindow *Wnd, int MoveFocused)
{
	CWindow *OldFocus;

	if (FocusEntry)
		OldFocus = FocusEntry->Wnd;
	else
		OldFocus = NULL;

	if (!Wnd) {
		FocusEntry = NULL;
		return;
	}

	if (MoveFocused) {
		Remove(Wnd);
		Add(Wnd);
		FocusEntry = LastNode;
	}
	else
		FocusEntry = Locate(Wnd);

	SwitchFocus(OldFocus,Wnd);
}

CWindow *CWindowList::GetFocusWnd()
{
	return FocusEntry ? FocusEntry->Wnd : NULL;
}

void CWindowList::FocusNext()
{
	CWindow *OldFocus;

	if (!FocusEntry)
		return;
	OldFocus = FocusEntry->Wnd;

	do
		if (FocusEntry->Next)
			FocusEntry = FocusEntry->Next;
		else
			FocusEntry = WindowList->Next;
	while (!FocusEntry->Wnd->Enabled || !FocusEntry->Wnd->Visible || !FocusEntry->Wnd->FocusWnd);

	SwitchFocus(OldFocus,FocusEntry->Wnd);
}

CWindow *CWindowList::GetLastFocus()
{
	return LastFocus;
}

void CWindowList::FocusPrev()
{
	CWindow *OldFocus;

	if (!FocusEntry)
		return;
	OldFocus = FocusEntry->Wnd;

	do
		if (FocusEntry->Prev != WindowList)
			FocusEntry = FocusEntry->Prev;
		else
			FocusEntry = LastNode;
	while (!FocusEntry->Wnd->Enabled || !FocusEntry->Wnd->Visible || !FocusEntry->Wnd->FocusWnd);

	SwitchFocus(OldFocus,FocusEntry->Wnd);
}

PWndNode CWindowList::Locate(CWindow *Wnd)
{
	PWndNode Entry;

	for (Entry = WindowList->Next; Entry && Entry->Wnd != Wnd; Entry = Entry->Next);
	return Entry;
}

void CWindowList::FixDamage(long Left, long Top, long Width, long Height)
{
	CWindow *Entry;
	long VPLeft, VPTop;
	long ClipLeft, ClipTop, ClipWidth, ClipHeight;
	long lLeft, lTop;

	if (!Visible || AdjustArea(Left,Top,Width,Height) == -1)
		return;
	lLeft = Left + this->Left;
	lTop = Top + this->Top;
	if (hScreen)
		hScreen->BeforeFix(lLeft, lTop, Width,Height);
	if (hForm)
		hForm->BeforeFix(lLeft, lTop, Width,Height);

	GotoFirst();
	Entry = GetThis();
	while (Entry) {
		Entry->FixDamage(Left,Top,Width,Height);
		Entry = GetNext();
	}
	if (hScreen)
		hScreen->AfterFix(lLeft, lTop, Width,Height);
	if (hForm)
		hForm->AfterFix(lLeft, lTop, Width,Height);

}

void CWindowList::SetHandler(CScreen *hScreen)
{
	this->hScreen = hScreen;
}


void CWindowList::SetHandler(CForm *hForm)
{
	this->hForm = hForm;
}

void CWindowList::SwitchFocus(CWindow *From, CWindow *To)
{
	LastFocus = From;
	if (From)
		From->Blur();
	To->Focus();
	Graph->FlushQueue();
	if (From)
		Graph->QFlushArea(Left + From->Left,Top + From->Top,
								From->Width,From->Height);
	Graph->QFlushArea(Left + To->Left,Top + To->Top,To->Width,To->Height);
	Graph->FlushQueue();
}

void CWindowList::MouseDown(int Left, int Top)
{
	PWndNode Entry;

	for (Entry = LastNode;
		  Entry != WindowList &&
		  Entry->Wnd->MouseDown(Left - this->Left,Top - this->Top) == -1;
		  Entry = Entry->Prev);
}

void CWindowList::MouseUp()
{
	if (FocusEntry)
		FocusEntry->Wnd->MouseUp();
}

void CWindowList::MouseMove(int Left, int Top)
{
	if (FocusEntry)
		FocusEntry->Wnd->MouseMove(Left - this->Left,Top - this->Top);
}

int CWindowList::AdjustArea(long &iLeft, long &iTop, long &iWidth, long &iHeight)
{
	long iRight, iBottom;
	long Right, Bottom;

	iRight = iLeft + iWidth - 1;
	iBottom = iTop + iHeight - 1;
	if (iLeft >= Width || iRight < 0 || iTop >= Height || iBottom < 0)
		return -1;
	if (iLeft < 0) {
		iWidth += iLeft;
		iLeft = 0;
	}
	if (iTop < 0) {
		iHeight += iTop;
		iTop = 0;
	}
	if (iRight >= Width)
		iWidth -= iRight - Width + 1;
	if (iBottom >= Height)
		iHeight -= iBottom - Height + 1;

	return 0;
}
