/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>

#define Align(x,n)  ( ((x)+(n)-1) & (~((n)-(1))) )

#ifndef NULL
#  define NULL 0
#endif

typedef struct SMemDesc TMemDesc;
struct SMemDesc {
	unsigned short PageCount;
	unsigned short Reserved;
	union {
		// is in FreeList ? NextFree : PrevFree;
		TMemDesc *NextFree;
		TMemDesc *PrevFree;
	};
	TMemDesc *PrevUsed;
	TMemDesc *NextUsed;
};

extern TMemDesc UsedList;
TMemDesc FreeList = {0,0,NULL,NULL,&UsedList};
TMemDesc UsedList = {0,0,&FreeList,NULL,NULL};


void InitHeap(TMemDesc *HeapStart, unsigned short PageCount)
{
	HeapStart->PageCount = PageCount;
	HeapStart->NextFree = NULL;
	HeapStart->NextUsed = NULL;
	HeapStart->PrevUsed = &UsedList;
	FreeList.NextFree = HeapStart;
}

static void InitPrevFree(TMemDesc *UsedSubList, TMemDesc *PrevFree)
{
	for (; UsedSubList; UsedSubList = UsedSubList->NextUsed)
		UsedSubList->PrevFree = PrevFree;
}


//void *operator new (unsigned Size)
void *mem_alloc(unsigned Size)
{
	unsigned long PageCount;
	TMemDesc *Entry;
	TMemDesc *FreePart;
	TMemDesc *UsedSubList;

	Size = Align(Size,16) + 16;
	PageCount = Size >> 4;

	for (Entry = FreeList.NextFree;
		  Entry && Entry->PageCount < PageCount;
		  Entry = Entry->NextFree);
	if (!Entry) {
		printf("malloc(): Out of memory\n");
		return NULL; // out of memory
	}
	if (PageCount != Entry->PageCount) {
		FreePart = (TMemDesc *)( (unsigned long)Entry + (PageCount << 16) );
		FreePart->PageCount = Entry->PageCount - PageCount;
		FreePart->NextFree = Entry->NextFree;
		FreePart->NextUsed = Entry->NextUsed;
		FreePart->PrevUsed = Entry;
		Entry->PrevUsed->PrevFree->NextFree = FreePart;
		Entry->PrevUsed->NextUsed = Entry;

		Entry->PageCount = PageCount;
		Entry->NextUsed = NULL;
		Entry->PrevFree = Entry->PrevUsed->PrevFree;
		UsedSubList = FreePart->NextUsed;
	}
	else {
		FreePart = Entry->PrevUsed->PrevFree;
		FreePart->NextFree = Entry->NextFree;
		Entry->PrevUsed->NextUsed = Entry;
		Entry->NextUsed->PrevUsed = Entry;
		UsedSubList = Entry;
	}
	InitPrevFree(UsedSubList,FreePart);

	return (void *)((char *)Entry + 0x00010000);
}

//void operator delete (void *Block)
void mem_free(void *Block)
{
	TMemDesc *Entry;
	TMemDesc *PrevFree;
	TMemDesc *NextFree;

	if (!Block) {
		printf("free(): NULL pointer\n");
		return; // Block == NULL
	}
	(char *)Block -= 0x00010000;
	for (Entry = FreeList.NextFree, PrevFree = &FreeList;
		  Entry && (unsigned long)Entry < (unsigned long)Block;
		  PrevFree = Entry, Entry = Entry->NextFree);

	for (Entry = PrevFree->NextUsed;
		  Entry && Entry != Block;
		  Entry = Entry->NextUsed);
	if (!Entry) {
		printf("free(): Invalid pointer\n");
		return; // Block is invalid pointer
	}
	NextFree = PrevFree->NextFree;
	if (Entry->PrevUsed)
		Entry->PrevUsed->NextUsed = NULL;
	else {
		PrevFree->PageCount += Entry->PageCount;
		PrevFree->NextUsed = Entry->NextUsed;
		Entry = PrevFree;
	}
	if (Entry->NextUsed) {
		Entry->NextUsed->PrevUsed = NULL;
		if (Entry != PrevFree)
			InitPrevFree(Entry->NextUsed,Entry);
	}
	else
		if (NextFree) {
			Entry->PageCount += NextFree->PageCount;
			Entry->NextFree = NextFree->NextFree;
			Entry->NextUsed = NextFree->NextUsed;
			NextFree = Entry;
			InitPrevFree(Entry->NextUsed,Entry);
		}
		else
			Entry->NextFree = NULL;
	if (Entry != NextFree)
		Entry->NextFree = NextFree;
	if (Entry != PrevFree)
		PrevFree->NextFree = Entry;
}

unsigned long CoreLeft()
{
	TMemDesc *FreeEntry;
	unsigned short PageCount;

	for (PageCount = 0, FreeEntry = FreeList.NextFree;
		  FreeEntry; FreeEntry = FreeEntry->NextFree)
		PageCount += FreeEntry->PageCount;
	return (unsigned long)PageCount << 4;
}

char Heap[32768];

void main()
{
	void *PtrList[16];
	int Index;
	int Iteration;
	int MemSize;

	clrscr();
	for (Index = 0; Index < 16; ++Index)
		PtrList[Index] = NULL;

	randomize();

	InitHeap((TMemDesc *)Heap,2048);

	printf("No iterations, CoreLeft = %lu\n",CoreLeft());
	for (Iteration = 0; Iteration < 256; ++Iteration) {
		Index = random(16);
		mem_free(PtrList[Index]);
		MemSize = random(256);
		PtrList[Index] = mem_alloc(MemSize);
		printf("Iteration: %d, Entry = %d, MemSize = %d, CoreLeft = %lu\n",Iteration,Index,MemSize,CoreLeft());
	}
	for (Index = 0; Index < 16; ++Index)
		mem_free(PtrList[Index]);


	printf("All memory freed, CoreLeft = %lu\n",CoreLeft());
}




