/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <items.h>
#include <mem.h>
#include <disk.h>
#include <key.h>

static const char *FileName = "BOOTITEMXDF";

CBootItems::CBootItems(CFileSystem *FileSystemToUse, CPartList *PartListToUse, CKeyboard *KeyboardToUse)
{
	FileSystem = FileSystemToUse;
	PartList = PartListToUse;
	Keyboard = KeyboardToUse;
	ItemFile = new TItemFile;
	FATFix = new CFATFix;
	Reset();
	LastConfig = new CLastConfig(FileSystemToUse,PartListToUse);
}

CBootItems::~CBootItems()
{
	delete LastConfig;
}

TBootItem *CBootItems::GetBootItem(int Index)
{
	if (Index < 0 || Index >= ItemFile->BootItemCount)
		return NULL;
	return &ItemFile->BootItems[Index];
}

void CBootItems::Add(TBootItem &NewItem)
{
	Insert(ItemFile->BootItemCount,NewItem);
}

void CBootItems::Insert(int Index, TBootItem &NewItem)
{
	if (Index > ItemFile->BootItemCount)
		Index = ItemFile->BootItemCount;
	ItemsMoveDown(Index);
	memcpy(&ItemFile->BootItems[Index],&NewItem,sizeof (TBootItem));
	++ItemFile->BootItemCount;
}

int CBootItems::GetCount()
{
	return ItemFile->BootItemCount;
}

void CBootItems::Remove(int Index)
{
	if (Index >= ItemFile->BootItemCount)
		return;
	if (Index == ItemFile->InstantBoot)
		ItemFile->InstantBoot = -1;
	else
		if (Index < ItemFile->InstantBoot)
			--ItemFile->InstantBoot;

	ItemsMoveUp(Index + 1);
	--ItemFile->BootItemCount;
}

void CBootItems::Clear()
{
	ItemFile->BootItemCount = 0;
	ItemFile->InstantBoot = -1;
}

void CBootItems::MoveUp(int Index)
{
	TBootItem Entry;

	if (Index <= 0 || Index >= ItemFile->BootItemCount)
		return;
	memcpy(&Entry,&ItemFile->BootItems[Index],sizeof (TBootItem));
	memcpy(&ItemFile->BootItems[Index],&ItemFile->BootItems[Index - 1],sizeof (TBootItem));
	memcpy(&ItemFile->BootItems[Index - 1],&Entry,sizeof (TBootItem));
	if (ItemFile->InstantBoot == Index)
		--ItemFile->InstantBoot;
	else
		if (ItemFile->InstantBoot == Index - 1)
			++ItemFile->InstantBoot;
}

void CBootItems::MoveDown(int Index)
{
	TBootItem Entry;

	if (Index < 0 || Index >= ItemFile->BootItemCount - 1)
		return;
	memcpy(&Entry,&ItemFile->BootItems[Index],sizeof (TBootItem));
	memcpy(&ItemFile->BootItems[Index],&ItemFile->BootItems[Index + 1],sizeof (TBootItem));
	memcpy(&ItemFile->BootItems[Index + 1],&Entry,sizeof (TBootItem));
	if (ItemFile->InstantBoot == Index)
		++ItemFile->InstantBoot;
	else
		if (ItemFile->InstantBoot == Index + 1)
			--ItemFile->InstantBoot;

}

void CBootItems::Save()
{
	FileSystem->WriteFile(FileName,ItemFile);
}

void CBootItems::Reset()
{
	FileSystem->ReadFile(FileName,ItemFile);
	if (ItemFile->InstantBoot >= ItemFile->BootItemCount)
		ItemFile->InstantBoot = -1;
}

void CBootItems::ItemsMoveUp(int Index)
{
	memcpy(&ItemFile->BootItems[Index - 1],
			 &ItemFile->BootItems[Index],
			 (ItemFile->BootItemCount - Index) * sizeof (TBootItem)
			);
}

void CBootItems::ItemsMoveDown(int Index)
{
	int MoveIndex;

	for (MoveIndex = ItemFile->BootItemCount; MoveIndex > Index; --MoveIndex)
		memcpy(&ItemFile->BootItems[Index],
				 &ItemFile->BootItems[Index - 1],sizeof (TBootItem));
}

int CBootItems::GetInstantBoot()
{
	return ItemFile->InstantBoot;
}

void CBootItems::SetInstantBoot(int Index)
{
	ItemFile->InstantBoot = Index;
}

int CBootItems::GetTimeout()
{
	return ItemFile->Timeout;
}

void CBootItems::SetTimeout(int Timeout)
{
	ItemFile->Timeout = Timeout;
}

int CBootItems::GetPartCount()
{
	return ItemFile->PartCount;
}

void CBootItems::DeinitBoot()
{
	UpdateBootItems(LastConfig->GetRelocTable());
	LastConfig->SetNewConfig();
}

int CBootItems::InitBoot(int BootIndex)
{
	CDisk Disk;
	int Index, Count;
	const TBootItem *BootItem;
	const TPartition *Partition;

	BootItem = &ItemFile->BootItems[BootIndex];
	Count = PartList->GetCount();
	Partition = PartList->GetPartition(BootItem->PartIndex);
	for (Index = 0; Index < Count; ++Index)
		switch (BootItem->HidingItems[Index]) {
			case ITEM_HIDE:
				PartList->Hide(Index);
				break;
			case ITEM_REVEAL:
				PartList->Reveal(Index);
				break;
			case ITEM_IGNORE:
			default:
				break;
		}
	Disk.Map(Partition->Drive,Partition->StartSector);
	if (Disk.Read(0,IPL_ADDR,1) == -1)
		return -1;
	if (BootItem->FixDriveNum)
		FATFix->Fix(Partition,(TBootRecord *)IPL_ADDR);
	if (BootItem->Activate)
		PartList->SetActive(BootItem->PartIndex);
	PartList->WriteStructure();
	StoreKeys(BootItem->Keys);
	return Partition->Drive;
}

void CBootItems::UpdateBootItems(const short *RelocTable)
{
	int Index;
	char NewHidingItems[92];
	int PartCount;
	int HideIndex;
	TBootItem *BootItem;

	PartCount = ItemFile->PartCount;
	memset(NewHidingItems,ITEM_IGNORE,sizeof (char[92]));
	for (Index = 0; Index < ItemFile->BootItemCount; ++Index) {
		BootItem = &ItemFile->BootItems[Index];
		if (BootItem->PartIndex != -1)
			if ((BootItem->PartIndex = RelocTable[BootItem->PartIndex]) == -1)
				BootItem->Valid = FALSE;
			else
				BootItem->Valid = TRUE;

		for (HideIndex = 0; HideIndex < PartCount; ++HideIndex)
			if (RelocTable[HideIndex] != -1)
				NewHidingItems[RelocTable[HideIndex]] = BootItem->HidingItems[HideIndex];
		memcpy(BootItem->HidingItems, NewHidingItems,PartCount);
	}
	ItemFile->PartCount = PartList->GetCount();
	Save();
}

void CBootItems::StoreKeys(const unsigned char *Keys)
{
	int Key;
	static const int CursorKeys[4] = {
		KEY_UP,KEY_DOWN,KEY_LEFT,KEY_RIGHT
	};

	Keyboard->Flush();
	while (*Keys) {
		Key = *Keys++;
		if (Key >= 244)
			Key = KEY_F1 + ((Key - 244) << 8);
		if (Key >= 191 && Key <= 194)
			Key = CursorKeys[Key - 191];
		Keyboard->StoreCh(Key);
	}
}

int CBootItems::CanFix(const TBootItem *Entry)
{
	const TPartition *Partition;

	Partition = PartList->GetPartition(Entry->PartIndex);
	return FATFix->CanFix(Partition);
}
