/*
 * Extended Operating System Loader (XOSL)
 * Copyright (c) 1999 by Geurt Vos
 *
 * This code is distributed under GNU General Public License (GPL)
 *
 * The full text of the license can be found in the GPL.TXT file,
 * or at http://www.gnu.org
 */

#include <xoslapp.h>
#include <cstring.h>
#include <text.h>
#include <timer.h>
#include <mem.h>

#include <mouse.h>
#include <graph.h>
#include <key.h>
#include <screen.h>
#include <fs.h>
#include <fat16.h>
#include <fat32.h>
#include <quit.h>

#include <ptab.h>
#include <xosldata.h>
#include <items.h>

#include <loader.h>
#include <setup.h>
#include <prefer.h>
#include <about.h>
#include <dialogs.h>
#include <rawboot.h>

#include <dump.h>

void printf(const char *format, ...);
long CoreLeft();

/**/

typedef struct {
	int Drive;
	int FSType;
	long StartSector;
} TMountPart;

#ifndef DOS_DEBUG

#define MOUNT_PART  ( *(const TMountPart *)0x00007c00 )

#else

//const TMountPart MOUNT_PART = {0x80,0x06,0x0031b600};

const TMountPart MOUNT_PART = {0x80,0x06,0x00246c00};
//const TMountPart MOUNT_PART = {0x80,0x06,0x005ffa40};
static char *FileName = "dump_a.dmp";

#endif


/**/

#define LOGO_WIDTH 218
#define LOGO_HEIGHT 146

#define WALLPAPER_WIDTH 150
#define WALLPAPER_HEIGHT 150

#define Int19Reboot() (__emit__(0xcd,0x19))

#define BypassRequest ((peekb(0x0000,0x0418) & 0x03) == 0x03 && \
							  (peekb(0x0000,0x0496) & 0x0c) == 0x0c)


/* Global vars. */
unsigned char *SplashLogo;
unsigned char *LogoBitmap;
unsigned char *WallpaperBitmap;
void *DefaultFont;
void *AdditionalFont;



/**/
const char *DefaultFontFileName = "SMLFNT10XFF";
const char *AddFontFileName     = "SMLFNT9 XFF";
const char *LogoFileName        = "XOSLLOGOXBF";
const char *SplashFileName      = "SPLASHLGXBF";
const char *WallpaperFileName   = "XOSLWALLXBF";

CApplication::CApplication()
{
	TextCapture = new CTextCapture;
}

CApplication::~CApplication()
{
	delete TextCapture;
}

int CApplication::Execute()
/*
 * XOSL main application function
 */
{
	int Key;
	int X, Y;
	int Status;
	TMiscPref *MiscPref;
	TBootItem *BootItem;

	if (BypassRequest)
		ExecuteBypass();
	CreateCoreClasses();
	InitializeCore();

	// reveal all partitions if requested
	MiscPref = XOSLData->GetMiscPref();
	if (MiscPref->AutoReveal)
		PartList->RevealAll();

	CreateApplication();

	DoSplashFadeOut();

	InitAppGraphics();

	/* Display main window */
	Loader->Show();
	Screen->InitScreen();

	DoAppFadeIn();


	Key = HandleInstantBoot();
	/* Main loop */
	do {
		while (Key != -1) {
			while (!Keyboard->KbHit() && !Loader->CanBoot()) {
				Mouse->GetXY(X,Y);
				Graph->SetCursorXY(X,Y);
				Screen->MouseStatus(X,Y,Mouse->MouseDown());
			}
			if (Loader->CanBoot())
				Key = -1;
			else {
				Key = Keyboard->GetCh();
				Screen->KeyPress(Key);
				if (!Preference->GetIgnoreNextKey()) {
					if (Key == MiscPref->KeyCycle)
						CycleWindow();
					else
#ifndef DOS_DEBUG
						if (Key == MiscPref->KeyShutdown)
							Shutdown();
						else
							if (Key == MiscPref->KeyRestart)
								Restart();
							else
								if (Key == MiscPref->KeyColdReboot)
									Reboot();
#else
						ScanForDebugKeys(Key);
#endif
				}
			}
		}

#ifdef DOS_DEBUG
		if (MiscPref->SetInstBoot) {
			BootItems->SetInstantBoot(Loader->GetBootItemIndex());
			BootItems->Save();
		}
		/* terminate program instead of booting */
		delete Mouse;
		Graph->SetMode(modeText);
		puts(BootItems->GetBootItem(Loader->GetBootItemIndex())->ItemName);
		asm mov ah,0x4c
		asm int 0x21
#endif

		PartList->SetAllowActiveHD(MiscPref->ActiveAllow);
		Status = BootItems->InitBoot(Loader->GetBootItemIndex());
		if (Status == -1) {
			DisplayBootError();
			Loader->CancelBoot();
			Key = 0;
		}
	} while (Status == -1);

	if (MiscPref->SetInstBoot) {
		BootItems->SetInstantBoot(Loader->GetBootItemIndex());
		BootItems->Save();
	}

	if (!XOSLData->GetGraphics()->DisableFadeOut)
		Graph->Palette->FadeToBlack();
	Graph->SetMode(modeText);

	DestroyApplication();
	FinalizeCore();
	DestroyCoreClasses();
	return Status;
}

void CApplication::CriticalError(const char *ErrorMsg)
{
	int Key;

	Graph->SetMode(modeText);
	puts(ErrorMsg);
	puts("\r\n");
	delete Mouse; /* only class that really needs destruction */
	ExecuteBypass();
}

void CApplication::ExecuteBypass(/*int Password*/)
{
	int Key;

#ifndef DOS_DEBUG
	puts("\r\n-- XOSL Bypass -- \r\n\nEnter number of partition to boot (0-3): ");
	while ((Key = Keyboard->GetCh() & 0x00ff) < '0' || Key > '3');
	puts((char *)&Key);
	puts("\r\n\n");
	RawBoot(Key - '0');
#else
	asm mov ah,0x4c
	asm int 0x21
#endif

}

void CApplication::CreateCoreClasses()
{
	Mouse = new CMouse;
	Graph = new CGraph(Mouse);
	Keyboard = new CKeyboard;
	Screen = new CScreen;
	switch (MOUNT_PART.FSType) {
		case 0x06: /* FAT16 */
			FileSystem = new CFAT16;
			break;
		case 0x0b: /* FAT32 */
			FileSystem = new CFAT32;
			break;
		default:
			CriticalError("Unknown filesystem.");
			break;
	}
}

void CApplication::DestroyCoreClasses()
{
	delete FileSystem;
	delete Screen;
	delete Keyboard;
	delete Graph;
	delete Mouse;
}

void CApplication::InitializeCore()
{
	TGraphData *GraphData;

	/* Initialize splash screen */
	FileSystem->Mount(MOUNT_PART.Drive,MOUNT_PART.StartSector);
	LoadXOSLSplashLogo();
	Screen->SetSplashLogo(LOGO_WIDTH,LOGO_HEIGHT,SplashLogo);
	XOSLData = new CXOSLData(FileSystem);
	GraphData = XOSLData->GetGraphics();
	if (Screen->ShowSplashScreen(GraphData->VideoMode,GraphData->LinearFB) == -1)
		CriticalError("Graphics mode not supported.");
	/* Load additional data */
	LoadDefaultFont();
	Graph->SetFont(DefaultFont);
	LoadAdditionalFont();
	LoadXOSLLogo();
	LoadWallpaper();
	Screen->SetWallpaper(WALLPAPER_WIDTH,WALLPAPER_HEIGHT,WallpaperBitmap);
	PartList = new CPartList;
	BootItems = new CBootItems(FileSystem,PartList,Keyboard);
	/* Additional initialization */
	BootItems->DeinitBoot();
	InitializeMouse();
}

void CApplication::FinalizeCore()
{
	delete BootItems;
	delete XOSLData;
	delete PartList;
}

void CApplication::CreateApplication()
{
	// ignore screen updates
	Screen->SetIgnoreSU(TRUE);

	Loader = new CLoader;
	About = new CAbout(Loader,Screen);
	Setup = new CSetup(Loader,Screen,PartList,BootItems,Keyboard);
	Preference = new CPreference(Loader,Setup,About,Screen,XOSLData,Keyboard,Mouse);

	Dialogs = new CDialogs(Screen,NULL,this);

	Loader->Initialize(Preference,Setup,About,Screen,BootItems,XOSLData);
	Preference->InitializeData();
	Setup->InitializeData();
	
	Screen->SetIgnoreSU(FALSE);
}

void CApplication::DestroyApplication()
/*
 * delete takes time, so if not really necessary, then don't.
 */
{
/*	delete About;
	delete Preference;
	delete Setup;
	delete Loader;*/
}

void CApplication::LoadDefaultFont()
{
	DefaultFont = new char[8192];
	if (!FileSystem->ReadFile(DefaultFontFileName,DefaultFont))
		CriticalError("Font file not found.");
}

void CApplication::LoadXOSLSplashLogo()
{
	SplashLogo = new unsigned char [(unsigned int)(LOGO_WIDTH * LOGO_HEIGHT)];
	if (!FileSystem->ReadFile(SplashFileName,SplashLogo))
		CriticalError("Logo bitmap not found.");
}

void CApplication::LoadXOSLLogo()
{
	LogoBitmap = new unsigned char [(unsigned int)(LOGO_WIDTH * LOGO_HEIGHT)];
	if (!FileSystem->ReadFile(LogoFileName,LogoBitmap))
		CriticalError("Logo bitmap not found.");
}

void CApplication::LoadAdditionalFont()
{
	AdditionalFont = new char [8192];
	if (!FileSystem->ReadFile(AddFontFileName,AdditionalFont))
		CriticalError("Font file not found.");
}

void CApplication::LoadWallpaper()
{
	WallpaperBitmap = new unsigned char [(unsigned int)(WALLPAPER_WIDTH * WALLPAPER_HEIGHT)];
	if (!FileSystem->ReadFile(WallpaperFileName,WallpaperBitmap))
		CriticalError("Wallpaper bitmap not found.");
}

void CApplication::InitializeMouse()
{
	TMouseData *MouseData;
	int MousePort;

	/* Initialize mouse */
	MouseData = XOSLData->GetMouse();
	if (MouseData->MouseType != -1)
		if (Mouse->SetMouse(MouseData->MouseType) == -1) {
			for (MousePort = 0; MousePort < 5 && !Mouse->Detect(MousePort); ++MousePort);
			if (MousePort == 5) /* Mouse not found, so set to 'no mouse' */
				MouseData->MouseType = -1;
		}
}

void CApplication::InitAppGraphics()
{
	int Width, Height;

	Graph->GetModeMetrics(Width,Height);
	/* Initialize mouse window, etc.
		Doesn't matter if no mouse installed */
	Mouse->SetWindow(0,0,Width,Height);
	Width >>= 1;
	Height >>= 1;
	Mouse->SetXY(Width,Height);
	Graph->SetCursorXY(Width,Height);
	/* The next call will have no effect if no mouse installed */
	Graph->ShowCursor();
	/* Relocate all windows */
	Loader->CenterWindow();
	Setup->CenterWindow();
	Preference->CenterWindow();
	About->CenterWindow();
}

void CApplication::DisplayBootError()
{
	CString ErrMsg;

	ErrMsg = "Unable to boot '";
	ErrMsg += BootItems->GetBootItem(Loader->GetBootItemIndex())->ItemName;
	ErrMsg += "'.";
	Loader->ShowBootError(ErrMsg);
}

void CApplication::Shutdown()
{
	CQuit Quit(Mouse);

	About->Hide();
	Loader->Hide();
	Preference->Hide();
	Setup->Hide();
	Dialogs->ShowAlertDialog("Shutdown","It is now safe to turn off your computer","Reboot",(TAlertProc)ShutDownReboot);
	if (XOSLData->GetMiscPref()->SoftPowerOff)
		Quit.Shutdown();
}

void CApplication::ShutDownReboot(CApplication *Application)
{
	Application->Reboot();
}

void CApplication::Restart()
{
	CQuit Quit(Mouse);

	Quit.Restart();
}

void CApplication::Reboot()
{
	CQuit Quit(Mouse);

	Quit.Reboot();
}

void CApplication::DoSplashFadeOut()
{
	Graph->Palette->LockFading();
	if (!XOSLData->GetGraphics()->DisableFadeIn) {
		Graph->Palette->SetFadeOutColor(XOSLData->GetColor()->FadeColorInt);
		Graph->Palette->FadeOut();
	}
	else {
		Graph->Palette->SetFadeOutColor(0);
		Graph->Palette->UltraFadeOut();
	}
}

void CApplication::DoAppFadeIn()
{
	if (!XOSLData->GetGraphics()->DisableFadeIn)
		Graph->Palette->FadeIn();
	else
		Graph->Palette->UltraFadeIn();
}

int CApplication::HandleInstantBoot()
/*
 * When DOS_DEBUG is defined, the progress bar and boot string
 * stay visible after a key press/mouse click
 */
{
	int X, Y;
	unsigned long Time1, Time2, Timeout, TimePassed;
	long Progress;
	int Sec, Min;
	CString BootStr;
	TMiscPref *MiscPref;
	int ContinueIB = 1;

	Setup->SetTimerSlider(BootItems->GetTimeout());
	/* Handle Instant Boot here */
	if (BootItems->GetInstantBoot() == -1)
		return 0;

	if (!BootItems->GetBootItem(BootItems->GetInstantBoot())->Valid) {
		Loader->AbortInstantBoot();
		return 0;
	}

	MiscPref = XOSLData->GetMiscPref();

	Timeout = ((unsigned long)BootItems->GetTimeout() << 16) / 3600;
	TimePassed = 0;
	Time1 = GetTimerTicks();
	while (ContinueIB) {
		Time2 = GetTimerTicks();
		TimePassed = Time2 - Time1;
		Progress = TimePassed * 3600;
		Sec = BootItems->GetTimeout() - (Progress >> 16);
		Progress >>= 12;
		BootStr = "Booting in ";
		if (Sec >= 60) {
			Min = Sec / 60;
			Sec %= 60;
			BootStr += Min;
			if (Sec >= 10)
				BootStr += ".";
			else
				BootStr += ".0";
			BootStr += Sec;
			BootStr += " minutes.";
		}
		else {
			BootStr += Sec;
			BootStr += " seconds.";
		}
		Loader->SetTimeoutLabelCaption(BootStr);
		Loader->SetTimeoutProgress(Progress);
		Mouse->GetXY(X,Y);
		Graph->SetCursorXY(X,Y);

		ContinueIB = !Keyboard->KbHit() && TimePassed <= Timeout;
		if (!MiscPref->IgnoreMouseClick)
			ContinueIB = ContinueIB && !Mouse->MouseDown();
	}
#ifndef DOS_DEBUG
	Loader->AbortInstantBoot();
#endif
	if (!Keyboard->KbHit() && !Mouse->MouseDown())
		return -1;
	return 0;
}

void CApplication::CycleWindow()
/*
 * Loader - Setup - Preference - About
 */
{
	int VisibleWin[4];
	int Index;

	if (Loader->GotFocus())
		Index = 1;
	else
		if (Setup->GotFocus())
			Index = 2;
		else
			if (Preference->GotFocus())
				Index = 3;
			else
				if (About->GotFocus())
					Index = 0;
				else
					return; // Shutdown is pressed, without Soft power-off

	VisibleWin[0] = Loader->IsVisible();
	VisibleWin[1] = Setup->IsVisible();
	VisibleWin[2] = Preference->IsVisible();
	VisibleWin[3] = About->IsVisible();
	while (!VisibleWin[Index])
		Index = (Index + 1) & 3;
	switch (Index) {
		case 0:
			Loader->Show();
			break;
		case 1:
			Setup->Show();
			break;
		case 2:
			Preference->Show();
			break;
		case 3:
			About->Show();
			break;
	}
}


#ifdef DOS_DEBUG
void CApplication::ScanForDebugKeys(int Key)
/*
 * Keys for debug information
 * F9 - Terminate XOSL
 * F8 - Refresh screen
 * F7 - Set cursor position to 0,0
 * F6 - Screen shot
 * F5 - Dump RGB palette
 * F4 - Print CoreLeft()
 */
{
	if (Key == KEY_F9) {
		printf("delete Mouse\n");
		delete Mouse;
		printf("Graph->SetMode(modeText);\n");
		Graph->SetMode(modeText);
		printf(".exit\n");
		asm mov ah,0x4c
		asm int 0x21
	}
	if (Key == KEY_F8)
		Screen->Refresh();
	if (Key == KEY_F7)
		gotoxy(0,0);
	if (Key == KEY_F6) {
		DumpScreen(FileName);
		++FileName[5];
	}
	if (Key == KEY_F5)
		DumpPalette("pal.dmp");
	if (Key == KEY_F4) {
		printf("\nCoreLeft(): %ld\n",CoreLeft());
		gotoxy(0,0);
	}
}
#endif

